/******************************************************************************
 This file is a part od FDD3000 Emulator

 Copyright (C) 2013,2014 by Sławomir Szczyrba <sszczyrba@gmail.com>
 Copyright (C) 2014 by Stefano Bodrato <stefano_bodrato hotmail dt com>

 This program is free software: you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License as published by the Free
 Software Foundation, either version 2 of the License, or (at your option) any
 later version.
    
 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this program. If not, see <http://www.gnu.org/licenses/lgpl.html>
******************************************************************************/

//#define DEBUG 1
#undef DEBUG

#ifdef DEBUG
    #define MSG(format, ...) fprintf (stderr, format, ## __VA_ARGS__)
#else
    #define MSG(...)
#endif

#include <config.h>
#include <libspectrum.h>
#include "settings.h"
#include "ui/ui.h"
#include "module.h"
#include "unittests/unittests.h"

#include "ti_m397.h"
#include "ti_m397ipc.h"

/*************************************************************************/

libspectrum_byte tim397_io_read( libspectrum_word port, int *attached ) { 
    libspectrum_byte b = 0;
    if(ipc&&connected) {
//MSG("Reading from port: %u,",port&0xff);
        if(fddIsAlive()) { b=fddRead(); } else  { connected=0; /*semResetAll();*/ return 0; }
        *attached = 1;
    }
//MSG("%u\n",b&0xff);
    return b;
}

void tim397_io_write( libspectrum_word port, libspectrum_byte b ) {
    if(ipc&&connected) { 
//MSG("Writing %u to port: %u\n",b&0xff,port&0xff);
	if(fddIsAlive()) { fddWrite(b); } else  { connected=0; /*semResetAll();*/ }
    }
}

/*************************************************************************/

static void tim397_reset( int hard_reset );
static void tim397_memory_map( void );
static void tim397_enabled_snapshot( libspectrum_snap *snap );
static void tim397_from_snapshot( libspectrum_snap *snap );
static void tim397_to_snapshot( libspectrum_snap *snap );
static void tim397_activate( void );


/* Two 8Kb memory chunks accessible by the Z80 when /ROMCS is low */
static memory_page tim397_memory_map_romcs_rom[ MEMORY_PAGES_IN_8K ];
static memory_page tim397_memory_map_romcs_ram[ MEMORY_PAGES_IN_8K ];
static int tim397_memory_source;

int tim397_available = 0;
int tim397_active = 0;

static libspectrum_byte *tim397_ram;
static int memory_allocated = 0;

static module_info_t tim397_module_info = {
  tim397_reset,
  tim397_memory_map,
  tim397_enabled_snapshot,
  tim397_from_snapshot,
  tim397_to_snapshot
};

void
tim397_page( void )
{
	tim397_active = 1;
	machine_current->ram.romcs = 1;
	machine_current->memory_map();
	ui_statusbar_update( UI_STATUSBAR_ITEM_DISK, UI_STATUSBAR_STATE_ACTIVE );
	MSG("tim397_page\n");
}

void
tim397_unpage( void )
{
	tim397_active = 0;
	machine_current->ram.romcs = 0;
	machine_current->memory_map();
	ui_statusbar_update( UI_STATUSBAR_ITEM_DISK, UI_STATUSBAR_STATE_INACTIVE );
	MSG("tim397_unpage\n");
}

static void
tim397_memory_map( void )
{
  if( !tim397_active ) return;
  memory_map_romcs_8k( 0x0000, tim397_memory_map_romcs_rom );
  memory_map_romcs_8k( 0x2000, tim397_memory_map_romcs_ram );
  MSG("tim397_memory_map\n");
}

static const periph_port_t tim397_ports[] = {
  { 0x00ff, 0x00ef, tim397_io_read, tim397_io_write },
  { 0, 0, NULL, NULL }
};

static const periph_t tim397_periph = {
  &settings_current.ti_m397,
  tim397_ports,
  1,
  tim397_activate
};

int
tim397_init( void )
{
  int i;
  module_register( &tim397_module_info );

  tim397_memory_source = memory_source_register( "TI_M397" );
  for( i = 0; i < MEMORY_PAGES_IN_8K; i++ )
    tim397_memory_map_romcs_rom[ i ].source = tim397_memory_source;
  for( i = 0; i < MEMORY_PAGES_IN_8K; i++ )
    tim397_memory_map_romcs_ram[ i ].source = tim397_memory_source;
  periph_register( PERIPH_TYPE_TIM397, &tim397_periph );

  ipc=0;

  MSG("tim397_init\n");

  return 0;
}


static void
tim397_reset( int hard_reset )
{
  int i;

  tim397_active = 0;
  tim397_available = 0;

  if( !periph_is_active( PERIPH_TYPE_TIM397 ) ) {
    ui_statusbar_update( UI_STATUSBAR_ITEM_DISK,
                         UI_STATUSBAR_STATE_NOT_AVAILABLE );
    MSG("FDD3000 is inactive\n");
    return;
  }

  if (openShm()!=0) {
//    settings_current.ti_m397 = 0;
//    periph_activate_type( PERIPH_TYPE_TIM397, 0 );
    MSG("Link to FDD3000 emulator failed\n");
//    return;
  }

  if( machine_load_rom_bank( tim397_memory_map_romcs_rom, 0,
                             settings_current.rom_tim397,
                             settings_default.rom_tim397, 0x1000 ) ||
      machine_load_rom_bank( tim397_memory_map_romcs_rom, 0x1000,
                             settings_current.rom_tim397,
                             settings_default.rom_tim397, 0x1000 ) ) {
    settings_current.ti_m397 = 0;
    periph_activate_type( PERIPH_TYPE_TIM397, 0 );
    ui_statusbar_update( UI_STATUSBAR_ITEM_DISK,
                         UI_STATUSBAR_STATE_NOT_AVAILABLE );
    MSG("Cannot load ROM file (%s, nor %s)\n",settings_current.rom_tim397,settings_default.rom_tim397);
    return;
  }
  
  machine_current->ram.romcs = 0;
  
  for( i = 0; i < MEMORY_PAGES_IN_8K; i++ ) {
	tim397_memory_map_romcs_ram[ i ].page = &tim397_ram[ i * MEMORY_PAGE_SIZE ];
	tim397_memory_map_romcs_ram[ i ].writable = 1;
  }

  tim397_available = 1;
  tim397_active = 0;

  if( hard_reset ) {
	memset( tim397_ram, 0, 0x2000 );
  }
 
  machine_current->memory_map();
  
  ui_statusbar_update( UI_STATUSBAR_ITEM_DISK, UI_STATUSBAR_STATE_INACTIVE );
  MSG("tim397_reset\n");
}

void
tim397_end( void )
{
    tim397_available = 0;
    closeShm(); // ? //
    cleanUp();
    MSG("tim397_end\n");
}

static void
tim397_activate( void )
{
  if( !memory_allocated ) {
    tim397_ram = memory_pool_allocate_persistent( 0x2000, 1 );
    memory_allocated = 1;
  }
  MSG("tim397_activate\n");
}

int
tim397_unittest( void )
{
  int r = 0;

  tim397_page();

  r += unittests_assert_16k_page( 0x0000, tim397_memory_source, 0 );
  r += unittests_assert_16k_ram_page( 0x4000, 5 );
  r += unittests_assert_16k_ram_page( 0x8000, 2 );
  r += unittests_assert_16k_ram_page( 0xc000, 0 );

  tim397_unpage();

  r += unittests_paging_test_48( 2 );

  return r;
}


static void
tim397_enabled_snapshot( libspectrum_snap *snap )
{
  return;
}

static void
tim397_from_snapshot( libspectrum_snap *snap )
{
  return;
}

static void
tim397_to_snapshot( libspectrum_snap *snap GCC_UNUSED )
{
  return;
}

